package ru.dbotthepony.kstarbound.world.entities

import it.unimi.dsi.fastutil.bytes.ByteArrayList
import ru.dbotthepony.kommons.util.MailboxExecutorService
import ru.dbotthepony.kommons.vector.Vector2d
import ru.dbotthepony.kstarbound.client.StarboundClient
import ru.dbotthepony.kstarbound.client.render.LayeredRenderer
import ru.dbotthepony.kstarbound.defs.JsonDriven
import ru.dbotthepony.kstarbound.world.Chunk
import ru.dbotthepony.kstarbound.world.ChunkPos
import ru.dbotthepony.kstarbound.world.LightCalculator
import ru.dbotthepony.kstarbound.world.World
import java.io.DataInputStream
import java.io.DataOutputStream
import java.util.UUID
import kotlin.concurrent.withLock

abstract class AbstractEntity(path: String) : JsonDriven(path) {
	/**
	 * The chunk this entity resides in
	 */
	var chunk: Chunk<*, *>? = null
		set(value) {
			if (innerWorld == null) {
				throw IllegalStateException("Trying to set chunk this entity belong to before spawning in world")
			} else if (value != null && innerWorld != value.world) {
				throw IllegalArgumentException("$this belongs to $innerWorld, $value belongs to ${value.world}")
			} else if (value == field) {
				return
			}

			val oldChunk = field
			field = value

			world.lock.withLock {
				if (oldChunk == null && value != null) {
					world.orphanedEntities.remove(this)
					value.addEntity(this)
				} else if (oldChunk != null && value == null) {
					world.orphanedEntities.add(this)
					oldChunk.removeEntity(this)
				} else if (oldChunk != null && value != null) {
					value.transferEntity(this, oldChunk)
				}
			}
		}

	abstract val chunkPos: ChunkPos
	abstract val position: Vector2d

	var entityID: Int = 0
		set(value) {
			if (field != 0)
				throw IllegalStateException("Already has Entity ID set (to $field)")

			field = value
		}

	var mailbox = MailboxExecutorService()
		private set

	private var innerWorld: World<*, *>? = null

	val world: World<*, *>
		get() = innerWorld ?: throw IllegalStateException("Not in world")

	val isSpawned: Boolean
		get() = innerWorld != null


	/**
	 * If set, then the entity will be discoverable by its unique id and will be
	 * indexed in the stored world.  Unique ids must be different across all
	 * entities in a single world.
	 */
	var uniqueID: String? = null
		protected set

	/**
	 * Whenever this entity should be removed when chunk containing it is being unloaded
	 *
	 * Returning false will also stop entity from being saved to disk, and render entity orphaned
	 * when chunk containing it will get unloaded
	 */
	open val isApplicableForUnloading: Boolean
		get() = true

	protected open fun onJoinWorld(world: World<*, *>) { }
	protected open fun onRemove(world: World<*, *>) { }

	abstract fun writeToNetwork(stream: DataOutputStream, isLegacy: Boolean)

	abstract fun readDelta(stream: ByteArrayList, interpolationTime: Double = 0.0, isLegacy: Boolean)

	fun joinWorld(world: World<*, *>) {
		if (innerWorld != null)
			throw IllegalStateException("Already spawned (in world $innerWorld)")

		if (entityID == 0)
			entityID = world.nextEntityID.incrementAndGet()

		world.ensureSameThread()

		check(!world.entities.containsKey(entityID)) { "Duplicate entity ID: $entityID" }

		if (mailbox.isShutdown)
			mailbox = MailboxExecutorService()

		innerWorld = world
		world.entities[entityID] = this
		world.orphanedEntities.add(this)
		onJoinWorld(world)
	}

	fun remove() {
		val world = innerWorld ?: throw IllegalStateException("Not in world")
		world.ensureSameThread()

		mailbox.shutdownNow()
		chunk = null
		check(world.entities.remove(entityID) == this) { "Tried to remove $this from $world, but removed something else!" }
		world.orphanedEntities.remove(this)
		onRemove(world)
		innerWorld = null
	}

	open val isRemote: Boolean
		get() = innerWorld?.isRemote ?: false

	fun think() {
		thinkShared()

		if (isRemote) {
			thinkRemote()
		} else {
			thinkLocal()
		}
	}

	protected open fun thinkShared() {
		mailbox.executeQueuedTasks()
	}

	protected open fun thinkRemote() {

	}

	protected open fun thinkLocal() {

	}

	open fun render(client: StarboundClient, layers: LayeredRenderer) {

	}

	open fun addLights(lightCalculator: LightCalculator, xOffset: Int, yOffset: Int) {

	}
}
